#
/*
 *		M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *				   R D C O N F . C
 *
 *				Douglas P. Kingston III
 *
 *				Part of the MDQS Daemon
 *
 */

#include <stdio.h>
#include "queue.h"

extern	char *Qconffile;
extern	char *MDQSuser;
extern	char *Sysmgr;

char	*strdup();

/*
 *			R D C O N F
 *
 *	Read and configure the global variables as specified
 */
rdconf( paramtab )
struct paramtab paramtab[];
{
	register struct paramtab *pp;
	char	linebuf[LINESIZ];
	int	filestate = 0;		/* 0=devs, 1=queues, 2=map */
	int	argcnt;
	char	**fieldp;
	FILE	*cf;

	/*
	 *	Copy initializations into malloc'd space so it
	 *	can later be free'd with impunity.
	 */
	for( pp = paramtab; pp->paramstr != 0; pp++ )
		if( pp->vartype == STRPVAR && pp->parmp != NULL )
			if( *pp->parmp )
				*pp->parmp = strdup( *pp->parmp );
	
	/*****
	 ***** there should be some kind of protection against
	 ***** against simultaneous access of the configuration file...
	 *****/
	if(( cf = fopen (Qconffile, "r")) == NULL )
		bomb ("Cannot open configuration file (%s: %s)",
			Qconffile, sys_errlist[errno]);

	while (fgets (linebuf, LINESIZ, cf) != NULL) {

		argcnt = parse (linebuf, &fieldp);

		/*  "Blank" line or bad line, probably a comment line */
		if (argcnt <= 0)
			continue;

		/*  Lines that start with a "-" indicate end of a section */
		if (**fieldp == '-') {
			filestate++;
			continue;
		}

		if( filestate == 0 ) {
			/*
			 *  Process system parameters
			 */
			if( argcnt != 2 ) {
				continue;
			}

			setparam( fieldp, paramtab );
		}
		if( filestate > 0 )
			break;
	}

	fclose( cf );
}

setparam( fp, paramtab )
char	**fp;
struct paramtab paramtab[];
{
	struct paramtab *pp;

	for( pp = paramtab; pp->paramstr != 0; pp++ )
		if( strcmp( *fp, pp->paramstr ) == SAME ) {
			switch( pp->vartype ) {
			case INTVAR:
				if( pp->setfunc != NULL )
					(*pp->setfunc)( fp );
				if( pp->parmp != NULL )
					*((int *)pp->parmp) = atoi( *++fp );
				break;

			case STRPVAR:
				if( pp->setfunc != NULL )
					(*pp->setfunc)( fp );
				if( pp->parmp != NULL )
					if( *pp->parmp )
						free( *pp->parmp );
					*pp->parmp = strdup( *++fp );
				break;
			}
			return;
		}
}
