/*
 *			N E T D A E M O N . C
 *
 *  Server MDQS daemon, for 4.1a/4.2 BSD by DPK at BRL, 1 Jan 82 (Sigh...)
 *
 *		R E V I S I O N   H I S T O R Y
 */

#include <stdio.h>
#include <signal.h>
#include "queue.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/wait.h>
#include <pwd.h>

extern	int errno;
extern	char *Tmpqdir;

struct sockaddr_in addr;

char	*MDQSlogin = "mdqs";
int	Maxconnections = 4;	/* Maximum simultaneous connections */
int	numconnections = 0;	/* Number of currently active connections */
int	debug = 0;		/* If nonzero, give verbose output on stderr */
int	mdqsport = 666;		/* Default MDQS service port */
char	errbuf[BUFSIZ];		/* Logging will be line buffered */
char	programid[40];		/* identification string for log entries */
char	*srvargv[MAXFIELDS];	/* Actual smtp server process, full path */
int	srvargc = 0;

struct paramtab paramtab[] = {
	"mdqsid",	STRPVAR,	&MDQSlogin,		NULL,
	0,		0,		0,			0
};

main (argc, argv)
int	argc;
char	**argv;
{
	register int	skt;
	int	pid;
	int	i;
	int	alarmtrap();
	char	*them;
	char	hostname[32];
	char	workarea[32];
	struct	passwd *pw;

	rdconf (paramtab);

	setbuf( stderr, errbuf );
	sprintf( programid, "mdqsdae(%5.5d): ", getpid());
	gethostname( hostname, sizeof(hostname));	/* current hostname */

	if (chdir (Tmpqdir) < 0) {
		log ("Can't chdir to %s (%s)", Tmpqdir, sys_errlist[errno]);
		exit (9);
	}
	if ((pw = getpwnam(MDQSlogin)) == NULL) {
		log ("Can't find MDQSlogin (%s) in /etc/passwd", MDQSlogin);
		exit (8);
	}

	/* Parse args; args will override configuration file... */
	arginit( argc, argv );

	/* Alarms will be used to avoid getting trapped in a wait call */
	signal( SIGALRM, alarmtrap );

	addr.sin_addr.s_addr = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(mdqsport);   /* swap bytes */

	/*
	 *  Create a socket to accept a MDQS connection on.
	 *  The doloop is due to some sort of IPC bug in 4.1a.
	 */
	if (debug) log("opening socket...");
	i = 60;
	do{
		skt = socket( AF_INET, SOCK_STREAM, 0, 0 );
		if( debug ) log( "socket returned %d", skt);
		if( skt < 0 )
			sleep( 5 );
	} while( skt < 0 && i-- > 0 );
	if( skt < 0 ) {
		log( "can't open socket (errno %d)", errno);
		exit (99);
	}
	if( debug ) log( "socket open on %d", skt );

	if (bind (skt, &addr, sizeof addr, 0) < 0) {
		log( "can't bind socket (errno %d)", errno);
		exit (98);
	}

	listen (skt, Maxconnections+1);

	while (1)
	{
		struct sockaddr_in rmtaddr;
		int	len_rmtaddr = sizeof rmtaddr;
		int	tmpskt;
		int	status;

		/*
		 *  Accept a connection.
		 */
		if(( tmpskt = accept( skt, &rmtaddr, &len_rmtaddr, 0 )) < 0) {
			if( debug ) log( "accept error (%d)", errno);
			sleep(1);
			continue;
		}

		/* We have a valid open connection, start a server... */
		if(( pid = fork()) < 0 ) {
			log( "could not fork (%d)", errno);
			close( tmpskt );
		} else if( pid == 0 ) {
			/*
			 *  Child
			 */
			register struct hostent *hp;

			hp = gethostbyaddr ( &rmtaddr.sin_addr,
				sizeof rmtaddr.sin_addr, AF_INET );
			if( hp == NULL ) {
				/* This should produce a dotted quad */
				sprintf( workarea, "%lx", them );
				hp->h_name = &workarea[0];
			}

			log( "%s started", hp->h_name );
			dup2( tmpskt, 0 );
			dup2( tmpskt, 1 );
			if( tmpskt > 1 )
				close( tmpskt );
			setgid(pw->pw_gid);
			setuid(pw->pw_uid);
			srvargv[srvargc++] = "-h";
			srvargv[srvargc++] = hp->h_name;
			srvargv[srvargc] = 0;
			execv (*srvargv, srvargv);
			log( "server exec error (%d)", errno);
			exit (99);
		}

		/*
		 *  Parent
		 */
		close( tmpskt );
		numconnections++;

		/*
		 *  This code collects ZOMBIES and implements load
		 *  limiting by staying in the do loop while the
		 *  Maxconnections active.
		 */
		while (wait3 (&status, numconnections < Maxconnections
		    ? WNOHANG : 0, 0) > 0)
			numconnections--;
	}
}

alarmtrap() {
	signal( SIGALRM, alarmtrap );
	return;
}

arginit( argc, argv )
int	argc;
char	**argv;
{
	register int	i;
	register char	*arg;

	if( argc == 1 ) {
		log( "Usage: netdae [-d] [-n #connections] [-p port] server [args]");
		exit( 1 );
	}

	for( i = 1; i < argc; i++ ) {
		if( *argv[i] != '-' )
			break;
		arg = argv[i];
		arg++;
		switch( *arg ) {
		case 'd':
			debug++;
			break;

		case 'n':
			if( ++i == argc ) {
				log( "missing number of connections" );
				exit( 99 );
			}
			Maxconnections = atoi( argv[i] );
			if( Maxconnections <= 0 ) {
				log( "Bad number of connections '%s'",
				    argv[i] );
				exit( 99 );
			}
			if (debug) log( "Maxconnection now %d", Maxconnections );
			break;

		case 'p':
			if( ++i == argc ) {
				log( "missing port number" );
				exit( 99 );
			}
			mdqsport = atoi( argv[i] );
			if( mdqsport <= 0 ) {
				log( "Bad port number '%s'", argv[i] );
				exit( 99 );
			}
			if (debug) log( "MDQS port now %d", mdqsport );
			break;
		}
	}

	if( i < argc ) {
		register int j = 0;

		if( access( argv[i], 01 ) < 0 )	{	/* execute privs? */
			log( "Cannot access server program '%s'", argv[i] );
			exit( 99 );
		}
		if( debug ) log( "server is '%s'", argv[i] );
		srvargv[j++] = argv[i++];
		while (argv[i])
			srvargv[j++] = argv[i++];	
		srvargv[j] = 0;
		srvargc = j;
	} else {
		log( "MDQSserver program not specified!" );
		exit( 99 );
	}
}

/* VARARGS */
log( fmt, a, b, c, d )
char *fmt;
{
	fputs( programid, stderr );
	fprintf( stderr, fmt, a, b, c, d );
	fputc( '\n', stderr );
	fflush( stderr );
}
