/* need vcatn(), readrm() functions, see vcatserver.c */
/*
 * Cat Simulator for Versatec
 */ 

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <pwd.h>
#include "vp.h"

#define	onintr()	doflush()

/* set in vcopyn.c */
extern int	vpltmode[];
extern int	vprtmode[];

int	vformfeed[]	= ffcom;	/* form feed command */

#define FORMFEED	/* fanfold paper */
#define FORMPAT	0137500	/* special pattern to signal ff */

#ifdef	vax
#define	NFONTS		20
#else
#define NFONTS		4
#define	FONTSWAP
#endif

#define DISPATCHSIZE	256	/* must be a power of two */
#define DSIZ		((sizeof *dispatch)*DISPATCHSIZE)
#define SPECIALFONT	3
#define MAXF		4

#define CONVERT(n)	(n*(25./54.))
#define RECONVERT(n)	(n*(54./25.))

#define RASTER_LENGTH	2112	/* bits per raster line on device */
#define BYTES_PER_LINE	(RASTER_LENGTH/8)
#define NLINES		85
#define TROFF_BYTES_PER_LINE	(2112/8) /* was 1536/8 for 8.5" wide */
#define BUFFER_SIZE	(NLINES*TROFF_BYTES_PER_LINE)

char	buffer[BUFFER_SIZE];	/*  Big line buffers  */
char	*buf0p = &buffer[0];	/*  Zero origin in circular buffer  */

#if (TROFF_BYTES_PER_LINE != BYTES_PER_LINE)
#define DOUBLE_BUF
#endif

char	*calloc();
char	*allpanic();
struct	passwd *getpwuid();

#ifdef FONTSWAP
/*
 * FONTSWAP permits vcat to "swap" fonts to/from reserved areas
 * on disk in raw mode, rather than throwing away fonts after
 * they are no longer needed and then re-reading the font
 * through the system's normal i/o mechanism.  Since the fonts
 * are relatively large (10-60 blocks!), contiguous raw i/o
 * can be a big win.  Uses "/dev/bs[0-9]".
 */
#define ASSERT(exp,str)	if (!(exp)) {\
fprintf(stderr, "Vcat: Assertion botch: %s\n", str);\
onintr();\
}
#define	NSAVE	10
struct	fontsave {
	int fnum;
	int psize;
	int fontfd;
	int swapped;
} fontsave[NSAVE];

int	swpflg;
#endif

struct	header {
	short	magic;
	unsigned short 	size;
	short	maxx;
	short	maxy;
	short	xtnd;
} header;

struct	dispatch {
	unsigned short	addr;
	short	nbytes;
	char	up;
	char	down;
	char	left;
	char	right;
	short	width;
};

struct	fontdes {
	int    fnum;
	int    psize;
	struct	dispatch *disp;
	char   *bits;
} fontdes[NFONTS] = {
	-1,
	-1
};

struct	dispatch *dispatch;

struct	point_sizes {
	int	stupid_code;
	int	real_code;
} point_sizes[] = {
	010, 6,
	0, 7,
	01, 8,
	07, 9,
	02, 10,
	03, 11,
	04, 12,
	05, 14,
	0211, 16,
	06, 18,
	0212, 20,
	0213, 22,
	0214, 24,
	0215, 28,
	0216, 36,
	0, 0
};

int	lines;
int	vc;		/* versatec output file descriptor */
int	cfnum = -1;	/* current font type */
int	cpsize = 10;	/* current point size */
int	cfont = 1;	/* position of current font in fontdes table */
char	*bits;		/* pointer to real data of current font */
int	nfontnum = -1;	/* next font wanted */
int	fontwanted = 1;	/* is a new font wanted? */
int	npsize = 10;	/* next size wanted */
int	last_ssize = 02;/* stupid size codes */
int	xpos,ypos;	/* current location on the page */
int	back;		/* direction of escape (horizontal motions) left-right */
int	verd;		/* direction of lead (vertical motions) up-down */
int	esc;		/* amount of current escape motion */
int	lead;		/* amount of current lead motion */
int	mcase;		/* current character case */
int	railmag;	/* new font requested by input data */
double	row, col;
char	*fontname[MAXF];/* pointers to the font path names */
#define	MAXMAG	120
char	fnbuf[MAXMAG];	/* buffer to hold font path names */

char	asctab[128] = {
	'\0',	/*blank*/
	'h',	/*h*/
	't',	/*t*/
	'n',	/*n*/
	'm',	/*m*/
	'l',	/*l*/
	'i',	/*i*/
	'z',	/*z*/
	's',	/*s*/
	'd',	/*d*/
	'b',	/*b*/
	'x',	/*x*/
	'f',	/*f*/
	'j',	/*j*/
	'u',	/*u*/
	'k',	/*k*/
	'\0',	/*blank*/
	'p',	/*p*/
	'\06',	/*_ 3/4 em dash*/
	';',	/*;*/
	'\0',	/*blank*/
	'a',	/*a*/
	'\05',	/*rule*/
	'c',	/*c*/
	'`',	/*` open*/
	'e',	/*e*/
	'\'',	/*' close*/
	'o',	/*o*/
	'\021',	/*1/4*/
	'r',	/*r*/
	'\022',	/*1/2*/
	'v',	/*v*/
	'-',	/*- hyphen*/
	'w',	/*w*/
	'q',	/*q*/
	'/',	/*/*/
	'.',	/*.*/
	'g',	/*g*/
	'\023',	/*3/4*/
	',',	/*,*/
	'&',	/*&*/
	'y',	/*y*/
	'\0',	/*blank*/
	'%',	/*%*/
	'\0',	/*blank*/
	'Q',	/*Q*/
	'T',	/*T*/
	'O',	/*O*/
	'H',	/*H*/
	'N',	/*N*/
	'M',	/*M*/
	'L',	/*L*/
	'R',	/*R*/
	'G',	/*G*/
	'I',	/*I*/
	'P',	/*P*/
	'C',	/*C*/
	'V',	/*V*/
	'E',	/*E*/
	'Z',	/*Z*/
	'D',	/*D*/
	'B',	/*B*/
	'S',	/*S*/
	'Y',	/*Y*/
	'\0',	/*blank*/
	'F',	/*F*/
	'X',	/*X*/
	'A',	/*A*/
	'W',	/*W*/
	'J',	/*J*/
	'U',	/*U*/
	'K',	/*K*/
	'0',	/*0*/
	'1',	/*1*/
	'2',	/*2*/
	'3',	/*3*/
	'4',	/*4*/
	'5',	/*5*/
	'6',	/*6*/
	'7',	/*7*/
	'8',	/*8*/
	'9',	/*9*/
	'*',	/***/
	'\04',	/*minus*/
	'\01',	/*fi*/
	'\02',	/*fl*/
	'\03',	/*ff*/
	'\020',	/* cent sign */
	'\012',	/*ffl*/
	'\011',	/*ffi*/
	'(',	/*(*/
	')',	/*)*/
	'[',	/*[*/
	']',	/*]*/
	'\013',	/* degree */
	'\014',	/* dagger */
	'=',	/*=*/
	'\017',	/* registered */
	':',	/*:*/
	'+',	/*+*/
	'\0',	/*blank*/
	'!',	/*!*/
	'\07',	/* bullet */
	'?',	/*?*/
	'\015',	/*foot mark*/
	'|',	/*|*/
	'\0',	/*blank*/
	'\016',	/* copyright */
	'\010',	/* square */
	'$',	/*$*/
	'\0',
	'\0',
	'"',	/*"*/
	'#',	/*#*/
	'<',	/*<*/
	'>',	/*>*/
	'@',	/*@*/
	'\\',	/*\\*/
	'^',	/*^*/
	'{',	/*{*/
	'}',	/*}*/
	'~'	/*~*/
};

char	spectab[128] = {
	'\0',	/*blank*/
	'w',	/*psi*/
	'h',	/*theta*/
	'm',	/*nu*/
	'l',	/*mu*/
	'k',	/*lambda*/
	'i',	/*iota*/
	'f',	/*zeta*/
	'r',	/*sigma*/
	'd',	/*delta*/
	'b',	/*beta*/
	'n',	/*xi*/
	'g',	/*eta*/
	'u',	/*phi*/
	't',	/*upsilon*/
	'j',	/*kappa*/
	'\0',	/*blank*/
	'p',	/*pi*/
	'@',	/*at-sign*/
	'7',	/*down arrow*/
	'\0',	/*blank*/
	'a',	/*alpha*/
	'|',	/*or*/
	'v',	/*chi*/
	'"',	/*"*/
	'e',	/*epsilon*/
	'=',	/*=*/
	'o',	/*omicron*/
	'4',	/*left arrow*/
	'q',	/*rho*/
	'6',	/*up arrow*/
	's',	/*tau*/
	'_',	/*underrule*/
	'\\',	/*\*/
	'W',	/*Psi*/
	'\07',	/*bell system sign*/
	'\001',	/*infinity*/
	'c',	/*gamma*/
	'\002',	/*improper superset*/
	'\003',	/*proportional to*/
	'\004',	/*right hand*/
	'x',	/*omega*/
	'\0',	/*blank*/
	'(',	/*gradient*/
	'\0',	/*blank*/
	'U',	/*Phi*/
	'H',	/*Theta*/
	'X',	/*Omega*/
	'\005',	/*cup (union)*/
	'\006',	/*root en*/
	'\014',	/*terminal sigma*/
	'K',	/*Lambda*/
	'-',	/*minus*/
	'C',	/*Gamma*/
	'\015',	/*integral sign*/
	'P',	/*Pi*/
	'\032',	/*subset of*/
	'\033',	/*superset of*/
	'2',	/*approximates*/
	'y',	/*partial derivative*/
	'D',	/*Delta*/
	'\013',	/*square root*/
	'R',	/*Sigma*/
	'1',	/*approx =*/
	'\0',	/*blank*/
	'>',	/*>*/
	'N',	/*Xi*/
	'<',	/*<*/
	'\016',	/*slash (longer)*/
	'\034',	/*cap (intersection)*/
	'T',	/*Upsilon*/
	'\035',	/*not*/
	'\023',	/*right ceiling (rt of ")*/
	'\024',	/*left top (of big curly)*/
	'\017',	/*bold vertical*/
	'\030',	/*left center of big curly bracket*/
	'\025',	/*left bottom*/
	'\026',	/*right top*/
	'\031',	/*right center of big curly bracket*/
	'\027',	/*right bot*/
	'\021',	/*right floor (rb of ")*/
	'\020',	/*left floor (left bot of big sq bract)*/
	'\022',	/*left ceiling (lt of ")*/
	'*',	/*multiply*/
	'/',	/*divide*/
	'\010',	/*plus-minus*/
	'\011',	/*<=*/
	'\012',	/*>=*/
	'0',	/*identically equal*/
	'3',	/*not equal*/
	'{',	/*{*/
	'}',	/*}*/
	'\'',	/*' acute accent*/
	'\`',	/*` grave accent*/
	'^',	/*^*/
	'#',	/*sharp*/
	'\036',	/*left hand*/
	'\037',	/*member of*/
	'~',	/*~*/
	'z',	/*empty set*/
	'\0',	/*blank*/
	'Y',	/*dbl dagger*/
	'Z',	/*box rule*/
	'9',	/*asterisk*/
	'[',	/*improper subset*/
	']',	/*circle*/
	'\0',	/*blank*/
	'+',	/*eqn plus*/
	'5',	/*right arrow*/
	'8'	/*section mark*/
};

readrm( railmag )
char	*railmag;	/* name of the railmag file */
{
	register int i;
	register char *cp;
	register int rmfd;
	char c;

	if ((rmfd = open( railmag, 0)) < 0)
	{
		err("Vcat: can't open railmag file '%s'", railmag );
		exit(1);
	}
	cp = fnbuf;
	for (i=0; i<4; i++) {
		fontname[i] = cp;
		while (read(rmfd, &c, 1) == 1 && c != '\n') {
			*cp++ = c;
			if (cp >= &fnbuf[MAXMAG]) {
				err("Vcat: Railmag file too big");
				exit(1);
			}
		}
		*cp++ = '\0';
	}
	close(rmfd);
}

long
vcatn( df, llimit )
FILE	*df;		/* vcat input file */
long	llimit;		/* line limit */
{
	register int c;
	static int initialized;

	fflush( stdout );
	vc = fileno( stdout );
	ioctl( vc, SETSTATE, vpltmode );
	lines = col = esc = lead = ypos = verd = back = mcase = 0;
	row = 25;
	xpos = CONVERT(row);
	for (c = 0; c < sizeof buffer; c++)
		buffer[c] = 0;
	while ((c = getc(df)) != EOF) {
		if (!c)
			continue;
		if (c & 0200) {
			esc += (~c) & 0177;
			continue;
		}
		if (esc) {
			if (back)
				esc = -esc;
			col += esc;
			ypos = CONVERT(col);
			esc = 0;
		}
		if ((c & 0377) < 0100)	/*  Purely for efficiency  */
			goto normal_char;
		switch (c) {

		case 0100:
			if (initialized) {
#ifdef FORMFEED
				slop_lines(NLINES);
#ifdef DOUBLE_BUF
				flshbuf();
#endif
				formp();
				break;
#else
				goto out;
#endif
			}
			initialized++;
#ifndef	vax
			/*
			 * calloc(), at times, has problems allocating enuf
			 * memory for large point sizes, even though it's
			 * really available.
			 * (Reference "break bug" comment in malloc()).
			 * A way around this is to allocate and free a chunk
			 * of memory big enuf for the largest font right away,
			 * then when you really need it calloc will find it.
			 * Thanks to teklabs!rickl for this fix.
			 */
			{ char *mp;
			  int siz=32000;

			  while ((mp=calloc(siz,1)) == NULL)
				siz -= 1000;
			  free(mp);
			}
#endif
			railmag = 0;
			if (loadfont(railmag, cpsize) < 0)
				err("Vcat: Can't load initial font");
			break;

		case 0101:	/* lower rail */
			crail(railmag &= ~01);
			break;

		case 0102:	/* upper rail */
			crail(railmag |= 01);
			break;

		case 0103:	/* upper mag */
			crail(railmag |= 02);
			break;

		case 0104:	/* lower mag */
			crail(railmag &= ~02);
			break;

		case 0105:	/* lower case */
			mcase = 0;
			break;

		case 0106:	/* upper case */
			mcase = 0100;
			break;

		case 0107:	/* escape forward */
			back = 0;
			break;

		case 0110:	/* escape backwards */
			back = 1;
			break;

		case 0111:	/* stop */
			break;

		case 0112:	/* lead forward */
			verd = 0;
			break;

		case 0113:	/* undefined */
			break;

		case 0114:	/* lead backward */
			verd = 1;
			break;

		case 0115:	/* undefined */
		case 0116:
		case 0117:
			break;

		default:
			if ((c & 0340) == 0140) {  /* leading */
				lead = (~c) & 037;
				if (verd)
					lead = -lead;
#ifdef NOROT
				row += lead*3;	/* raw troff output */
#else
				row += lead;	/* rotated data */
#endif
				c = CONVERT(row);
				while (c >= NLINES) {
					slop_lines(15);
					c = CONVERT(row);
				}
				xpos = c;
				continue;
			}
			if ((c & 0360) == 0120) { /* size change */
				loadfont(railmag, findsize(c & 017));
				continue;
			}
			if (c & 0300)
				continue;
normal_char:
			c = (c & 077) | mcase;
			outc(c);
		}
	}
#ifndef FORMFEED
out:
#endif
	slop_lines(NLINES);
#ifdef DOUBLE_BUF
	flshbuf();
#endif
	ioctl(vc, SETSTATE, vprtmode);
}

findsize(code)
register int code;
{
	register struct point_sizes *psp;

	psp = point_sizes;
	while (psp->real_code != 0) {
		if ((psp->stupid_code & 017) == code)
			break;
		psp++;
	}
	code = 0;
	if (!(last_ssize & 0200) && (psp->stupid_code & 0200))
		code = -55;
	else if ((last_ssize & 0200) && !(psp->stupid_code & 0200))
		code = 55;
	if (back)
		code = -code;
#ifdef NOROT
	esc += code;	/* compensate for hardware shift during size change */
#endif
	last_ssize = psp->stupid_code;
	return(psp->real_code);
}

crail(nrail)
register int nrail;
{
	register int psize;

	psize = cpsize;
	if (fontwanted && psize != npsize)
		psize = npsize;
	loadfont(nrail, psize);
}

loadfont(fnum, size)
register int fnum;
register int size;
{
	register int i;

	fontwanted = 0;
	if (fnum == cfnum && size == cpsize)
		return(0);
	for (i = 0; i < NFONTS; i++)
		if (fontdes[i].fnum == fnum && fontdes[i].psize == size) {
			cfnum = fontdes[i].fnum;
			cpsize = fontdes[i].psize;
			dispatch = &fontdes[i].disp[0];
			bits = fontdes[i].bits;
			cfont = i;
			return(0);
		}
	if (fnum < 0 || fnum >= MAXF) {
		err("Vcat: Internal error (illegal font)");
		return(-1);
	}
	nfontnum = fnum;
	npsize = size;
	fontwanted++;
	return(0);
}

getvfont()
{
	register int fnum, size, font;
	int b, d;
	char cbuf[80];

	if (!fontwanted)
		return(0);
	fnum = nfontnum;
	size = npsize;
#ifdef FONTSWAP
	font = swapin(cbuf);
#else
	sprintf(cbuf, "%s.%d", fontname[fnum], size);
	font = open(cbuf, 0);
#endif
	if (font < 0) {
		fprintf(stderr, "Vcat: ");
		perror(cbuf);
		fontwanted = 0;
		return(-1);
	}
	if (read(font, &header, sizeof header)!=sizeof header || header.magic!=0436)
		err("Vcat: Bad font file: %s", cbuf);
	else {
		cfont = relfont();
		if (((bits=calloc(header.size+DSIZ+1, 1)) == NULL)
		    && ((bits=allpanic(header.size+DSIZ+1)) == NULL))
			err("Vcat: Not enuf memory for %s", cbuf);
		else {
			/*
			 * have allocated one chunk of mem for font, dispatch.
			 * get the dispatch addr, align to word boundary.
			 */
			d = (int)bits+header.size;
			d += 1;
			d &= ~1;
#ifdef FONTSWAP
			if (swpflg) {
				lseek(font, 512L, 0);
				b = read(font, bits, DSIZ + header.size);
			} else
#endif
			{
				b = read(font, d, DSIZ);
				b += read(font, bits, header.size);
			}
			if (b != DSIZ + header.size)
				err("Vcat: Bad font header: %s", cbuf);
			else {
#ifdef FONTSWAP
				if (!swpflg)
#endif
					close(font);
				cfnum = fontdes[cfont].fnum = fnum;
				cpsize = fontdes[cfont].psize = size;
				fontdes[cfont].bits = bits;
				fontdes[cfont].disp = (struct dispatch *)d;
				dispatch = &fontdes[cfont].disp[0];
				fontwanted = 0;
				return(0);
			}
		}
	}
#ifdef FONTSWAP
	ASSERT(swpflg==0, "good font went bad?");
#endif
	close(font);
	fontwanted = 0;
	return(-1);
}

int lastloaded = -1;

relfont()
{
	register int newfont;

	newfont = lastloaded;
	/*
	 * optimization for special font.  since we think that usually
	 * there is only one character at a time from any special math
	 * font, make it the candidate for removal.
	 */
	if (fontdes[cfont].fnum != SPECIALFONT || fontdes[cfont].bits==0)
		if (++newfont>=NFONTS)
			newfont = 0;
	lastloaded = newfont;
#ifdef FONTSWAP
	swapout(newfont);
#endif
	if ((int)fontdes[newfont].bits != -1 && fontdes[newfont].bits != 0)
		free(fontdes[newfont].bits);
	fontdes[newfont].bits = 0;
	return(newfont);
}

char *
allpanic(nbytes)
int nbytes;
{
	register int i;

	for (i = 0; i < NFONTS; i++)
		relfont();
	lastloaded = cfont;
	for (i = 0; i < NFONTS; i++) {
		fontdes[i].fnum = fontdes[i].psize = -1;
		fontdes[i].bits = 0;
		cfnum = cpsize = -1;
	}
	return(calloc(nbytes, 1));
}

#ifdef FONTSWAP
swapin(buf)
char *buf;
{
	register struct fontsave *p;
	register int fnum, size;
	static int initialized;

	swpflg = 0;
	fnum = nfontnum;
	size = npsize;
	if (!initialized) {
		initialized++;
		for (p=fontsave; p<&fontsave[NSAVE]; p++)
			p->fnum = -1;
	}
	else for (p=fontsave; p<&fontsave[NSAVE]; p++)
		if (p->fnum==fnum && p->psize==size) {
			ASSERT(p->swapped!=0, "font not swapped?");
			swpflg++;
			lseek(p->fontfd, 0L, 0);
			return(p->fontfd);
		}
	sprintf(buf, "%s.%d", fontname[fnum], size);
	return(open(buf, 0));
}

swapout(font)
register int font;
{
	register struct fontsave *p;
	char cbuf[80];
	int ffd, bsfd;
	char bsfile[20], *s;
	struct header hdr;

	if ((int)fontdes[font].bits != -1 && fontdes[font].bits != 0) {
		for (p=fontsave; p<&fontsave[NSAVE]; p++) {
			if (p->fnum == fontdes[font].fnum
			    && p->psize == fontdes[font].psize) {
				ASSERT(p->swapped!=0, "swap error");
				return;
			}
			if (p->fnum < 0)
				break;
		}
		if (p==&fontsave[NSAVE])
			return;
		strcpy(bsfile,"/dev/bsX");
		for (s=bsfile; *s!='X'; s++)
			;
		for (*s='0'; *s<='9'; (*s)++)
			if ((bsfd=open(bsfile,2)) > 0)
				break;
		if (bsfd < 0)
			return;	/* swap space exhausted */
		p->fnum = fontdes[font].fnum;
		p->psize = fontdes[font].psize;
		p->fontfd = bsfd;
		p->swapped = 1;
		sprintf(cbuf, "%s.%d", fontname[p->fnum], p->psize);
		ffd = open(cbuf, 0);
		read(ffd, &hdr, sizeof hdr);
		close(ffd);
		/*
		 * make size even for raw i/o and because
		 * dispatch table is aligned on word boundary
		 */
		hdr.size += 1;
		hdr.size &= ~1;
		write(bsfd, &hdr, sizeof hdr);
		lseek(bsfd, 512L, 0);
		write(bsfd, fontdes[font].bits, DSIZ + hdr.size);
	}
}
#endif

outc(code)
int code;
{
	char c;			/* character to print */
	struct dispatch *d;	/* ptr to character font record */
	register char *addr;	/* addr of font data */
	register int llen;	/* length of each font line */
	int nlines;		/* number of font lines */
	register char *scanp;	/* ptr to output buffer */
	int offset;		/* bit offset to start of font data */
	int i;			/* loop cointer */

	if (fontwanted && getvfont()<0)
		return(0);
	if (railmag == SPECIALFONT) {
		if ((c = spectab[code]) < 0)
			return(0);
	} else {
		if ((c = asctab[code]) < 0)
			return(0);
	}
	d = dispatch+c;
	if (d->nbytes) {
		addr = bits+d->addr;
		llen = (d->left+d->right+7)/8;
		nlines = d->up+d->down;
		if (xpos+d->down >= NLINES)
			slop_lines(xpos+d->down-NLINES+1);
		scanp = ((xpos-d->up-1)*TROFF_BYTES_PER_LINE+(ypos-d->left)/8)+buf0p;
		if (scanp < &buffer[0])
			scanp += sizeof buffer;
		offset = -((ypos-d->left)&07);
		for (i=0; i<nlines; i++) {
			if (scanp >= &buffer[BUFFER_SIZE])
				scanp -= sizeof buffer;
			stuff(addr, llen, scanp, offset);
			scanp += TROFF_BYTES_PER_LINE;
			addr += llen;
		}
		return(1);
	}
	return(0);
}

#ifdef FORMFEED
formp()
{
	ioctl(vc, SETSTATE, vformfeed );
	return;
}
#endif

#ifndef DOUBLE_BUF
slop_lines(nlines)
register int nlines;
{
	register int rlines, errval;

	lines += nlines;
	errval = 0;
	rlines = (&buffer[BUFFER_SIZE] - buf0p) / BYTES_PER_LINE;
	if (rlines < nlines) {
		errval = write(vc, buf0p, rlines * BYTES_PER_LINE);
		clear(buf0p, rlines * BYTES_PER_LINE);
		buf0p = buffer;
		nlines -= rlines;
		xpos -= rlines;
		row -= RECONVERT(rlines);
	}
	errval |= write(vc, buf0p, nlines * BYTES_PER_LINE);
	if (errval == -1) {
		err("Vcat: Versatec IO error");
		onintr();
	}
	clear(buf0p, nlines * BYTES_PER_LINE);
	buf0p += nlines * BYTES_PER_LINE;
	if (buf0p >= &buffer[BUFFER_SIZE])
		buf0p -= sizeof buffer;
	xpos -= nlines;
	row -= RECONVERT(nlines);
}

#else

/*
 * If troff does not put out a full plotter width of data,
 * we must have separate buffers for the font chars and the
 * actual output so that we can null pad the missing area.
 */
#define	BUFLGNTH	15
char	raster[BYTES_PER_LINE * BUFLGNTH];
char	*scanline = raster;
int	linecount;

slop_lines(nlines)
int nlines;
{
	register int *fp, *tp, i;
	int j;

	lines += nlines;
	fp = (int *)buf0p;
	for (j=0; j<nlines; j++) {
		if (fp >= &buffer[BUFFER_SIZE])
			fp -= sizeof(buffer)/sizeof(int);
		tp = (int *)scanline;
		i = TROFF_BYTES_PER_LINE/(sizeof *tp);
		do {
			*tp++ = *fp;
			*fp++ = 0;
		} while(--i);
		scanline += BYTES_PER_LINE;
		if (++linecount==BUFLGNTH)
			flshbuf();
	}
	buf0p += nlines * TROFF_BYTES_PER_LINE;
	if (buf0p >= &buffer[BUFFER_SIZE])
		buf0p -= sizeof buffer;
	xpos -= nlines;
	row -= RECONVERT(nlines);
}

flshbuf()
{
	if (write(vc, raster, linecount * BYTES_PER_LINE) == -1) {
		err("Vcat: Versatec IO error");
		onintr();
	}
	linecount = 0;
	scanline = raster;
}
#endif

#ifdef	vax
int	M[] = { 0xffffffff, 0xfefefefe, 0xfcfcfcfc, 0xf8f8f8f8,
		0xf0f0f0f0, 0xe0e0e0e0, 0xc0c0c0c0, 0x80808080, 0x0 };
int	N[] = { 0x00000000, 0x01010101, 0x03030303, 0x07070707,
		0x0f0f0f0f, 0x1f1f1f1f, 0x3f3f3f3f, 0x7f7f7f7f, 0xffffffff };
int	strim[] = { 0xffffffff, 0xffffff00, 0xffff0000, 0xff000000, 0 };

stuff(addr, count, scanp, offset)
register char *addr, *scanp;
register int count, offset;
{
	register unsigned fontdata;	/* font data temporary */
	register int off8;		/* offset+8 */

	off8 = offset+8;
	if (scanp + count <= &buffer[BUFFER_SIZE])
		do {
			fontdata = *(unsigned *)addr;
			addr += 4;
			if (count < 4)
				fontdata &= ~strim[count];
			*(unsigned *)scanp |= (fontdata << offset) & ~M[off8];
			scanp++;
			*(unsigned *)scanp |= (fontdata << off8) & ~N[off8];
			scanp += 3;
			count -= 4;
		} while (count > 0);
}

/* ARGSUSED */
clear(bp, nbytes)
int *bp;
int nbytes;
{
	asm("movc5 $0,(sp),$0,8(ap),*4(ap)");
}

#else

clear(bp, nbytes)
register int *bp;
register int nbytes;
{
	nbytes >>= 1;
	while (--nbytes>=0)
		*bp++ = 0;
}
#endif
